	/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2009 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2009.
 *   author Rub�n Blanco rblanco@prodevelop.es
 *
 *
 * Original version of the code made by Nicolas Gramlich.
 * No header license code found on the original source file.
 * 
 * Original source code downloaded from http://code.google.com/p/osmdroid/
 * package org.andnav.osm.views.util;
 * 
 * License stated in that website is 
 * http://www.gnu.org/licenses/lgpl.html
 * As no specific LGPL version was specified, LGPL license version is LGPL v2.1
 * is assumed.
 *  
 * 
 */

package es.prodevelop.tilecache.provider;

import java.util.logging.Level;
import java.util.logging.Logger;

import es.prodevelop.gvsig.mini.common.IBitmap;
import es.prodevelop.gvsig.mini.common.IContext;
import es.prodevelop.gvsig.mini.common.IHandler;
import es.prodevelop.gvsig.mini.common.impl.Tile;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.map.GeoUtils;
import es.prodevelop.gvsig.mini.utiles.Cancellable;
import es.prodevelop.gvsig.mini.utiles.WorkQueue;
import es.prodevelop.tilecache.draw.TileDrawer;
import es.prodevelop.tilecache.provider.event.TileEvent;
import es.prodevelop.tilecache.provider.filesystem.ITileFileSystemProvider;
import es.prodevelop.tilecache.provider.filesystem.impl.TileFilesystemProvider;
import es.prodevelop.tilecache.provider.filesystem.strategy.ITileFileSystemStrategy;
import es.prodevelop.tilecache.util.ConstantsTileCache;

/**
 * A class to manage tile downloading, saving/loading to/from disk and
 * putting/getting to/from memory cache
 * 
 * @author aromeu
 * @author rblanco
 * 
 */
public class TileProvider implements GeoUtils {

	/**
	 * The number of times that a tile is retried to download or load from fs.
	 * For a massive download process this can be set to a high value (for
	 * example 1000) to avoid failed tiles
	 */
	public static int NUM_RETRY = 2;

	/**
	 * Gets tiles from file system and from the server
	 */
	public static final int MODE_ONLINE = 0;

	/**
	 * Only gets tiles that are on file system
	 */
	public static final int MODE_OFFLINE = 1;

	/**
	 * If a tile is on the file system, deletes it and tries to download
	 */
	public static final int MODE_UPDATE = 2;

	/**
	 * Deletes the tile cache
	 */
	public static final int MODE_DELETE = 3;

	/**
	 * Don't save any tile, always download them
	 */
	public static final int MODE_CONNECTED = 4;

	/**
	 * Don't allocate bitmaps in memory. This mode is useful when there is no
	 * need to paint tiles on a canvas, for example when downloading tiles we
	 * don't need store them in memory, only store them on disk
	 */
	public boolean NOT_LOCAL_CACHE = false;

	public IBitmap mLoadingMapTile = null;
	public IBitmap mOfflineMapTile = null;
	protected IContext mCtx;
	String datalog = null;
	protected TileCache mTileCache;
	protected ITileFileSystemProvider mFSTileProvider;

	protected Downloader mTileDownloader;
	protected IHandler mLoadCallbackHandler;
	private final static Logger log = Logger.getLogger("TileProvider");
	protected int loadingTileID;
	protected int offlineTileID;
	private int mode;

	protected int lastZoomLevelRequested = -1;
	private Extent viewExtent;

	/**
	 * The constructor
	 * 
	 * @param ctx
	 *            A Context instance
	 * @param aDownloadFinishedListener
	 *            A Handler to send callbacks
	 * @param width
	 *            The map width to calculate memory cache size
	 * @param height
	 *            The map height to calculate memory cache size
	 * @param tileSize
	 *            A single tile size in pixels to calculate memory cache size
	 * @param mode
	 *            MODE_ONLINE = 0 - Gets tiles from file system and from the
	 *            server MODE_OFFLINE = 1 - Only gets tiles that are on file
	 *            system MODE_UPDATE = 2 - If a tile is on the file system,
	 *            deletes it and tries to download MODE_DELETE = 3 - Deletes the
	 *            tile cache MODE_CONNECTED = 4 - Don't save any tile, always
	 *            download them
	 * @param fileSystemStrategy
	 *            the ITileFileSystemStrategy
	 */
	public TileProvider(final IContext ctx,
			final IHandler aLoadCallbackHandler, final int width,
			final int height, final int tileSize, int mode,
			ITileFileSystemStrategy fileSystemStrategy) {
		try {
			// try {
			// CompatManager.getInstance().getRegisteredLogHandler().
			// configureLogger(log);
			// } catch (BaseException e) {
			//
			// }
			this.mode = mode;
			this.loadingTileID = ctx.getDefaultLoadingTileID();
			this.offlineTileID = ctx.getDefaultOfflineTileID();
			this.mLoadCallbackHandler = aLoadCallbackHandler;
			this.mCtx = ctx;
			WorkQueue.getInstance(ConstantsTileCache.DEFAULT_NUM_THREADS);
			try {
				this.mLoadingMapTile = this.mCtx
						.getBitmapFromResource(loadingTileID);
				this.mOfflineMapTile = this.mCtx
						.getBitmapFromResource(offlineTileID);
			} catch (OutOfMemoryError e) {
				System.gc();
				log.log(Level.SEVERE, "", e);
				this.onLowMemory();
			} catch (Exception e) {
				log.log(Level.SEVERE, "", e);
			}
			if (!NOT_LOCAL_CACHE)
				this.mTileCache = new TileCache(width, height, tileSize);
			this.mFSTileProvider = new TileFilesystemProvider(4 * 1024 * 1024,
					this.mTileCache, ctx, fileSystemStrategy, mode,
					this.mOfflineMapTile);
			this.mTileDownloader = new Downloader(ctx, this.mFSTileProvider);

		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
		}
	}
	
	/**
	 * The constructor
	 * 
	 * @param ctx
	 *            A Context instance
	 * @param aDownloadFinishedListener
	 *            A Handler to send callbacks
	 * @param width
	 *            The map width to calculate memory cache size
	 * @param height
	 *            The map height to calculate memory cache size
	 * @param tileSize
	 *            A single tile size in pixels to calculate memory cache size
	 * @param mode
	 *            MODE_ONLINE = 0 - Gets tiles from file system and from the
	 *            server MODE_OFFLINE = 1 - Only gets tiles that are on file
	 *            system MODE_UPDATE = 2 - If a tile is on the file system,
	 *            deletes it and tries to download MODE_DELETE = 3 - Deletes the
	 *            tile cache MODE_CONNECTED = 4 - Don't save any tile, always
	 *            download them
	 * @param notLocalCache if false not stores tiles in memory cache
	 * @param fileSystemStrategy
	 *            the ITileFileSystemStrategy
	 */
	public TileProvider(final IContext ctx,
			final IHandler aLoadCallbackHandler, final int width,
			final int height, final int tileSize, int mode,
			ITileFileSystemStrategy fileSystemStrategy, boolean notLocalCache) {
		this(ctx, aLoadCallbackHandler, width, height, tileSize, mode, fileSystemStrategy);
		try {
			this.NOT_LOCAL_CACHE = notLocalCache;
		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
		}
	}

	/**
	 * Clears the pending tile queues of Downloader and TileFilesystemProvider
	 * instances of this TileProvider. Also clear pending tasks on WorkQueue
	 */
	public void clearPendingQueue() {
		try {
			// log.log(Level.FINE, "clear pending queue");
			this.mFSTileProvider.getPendingQueue().clear();
			this.mTileDownloader.mPending.clear();
			WorkQueue.getInstance().clearPendingTasks();
			this.mCtx.sendMessage(this.mLoadCallbackHandler,
					TILEPROVIDER_CANCELED);
		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
		}
	}

	/**
	 * The strategy to get Tiles is: If the tile is in memory cache, then is
	 * returned and put the first (LRU strategy) If not then the tile is
	 * searched on the FileSystem if the tile is found, is put into the memory
	 * cache if not then is downloaded from the server, save to file system and
	 * finally put into memroy cache
	 * 
	 * @param tile
	 *            The tile containing the tile URL and the tile x-y, zoomLevel
	 *            and renderer name
	 * @param cancellable
	 *            A Cancellable instance
	 * @return A Bitmap representing the tile
	 */
	public IBitmap getMapTileFromCache(final Tile tile,
			final Cancellable cancellable) {
		try {
			// long t1 = System.currentTimeMillis();
			final String cacheURL = tile.getTileString();
			if (this.mTileCache == null) return null;
			final IBitmap b = this.mTileCache.getMapTile(cacheURL);
			if (b == null && mode == TileProvider.MODE_OFFLINE) {
				return this.mOfflineMapTile;
			}
			return b;
		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
			return this.mLoadingMapTile;
		}
	}

	/**
	 * The strategy to get Tiles is: If the tile is in memory cache, then is
	 * returned and put the first (LRU strategy) If not then the tile is
	 * searched on the FileSystem if the tile is found, is put into the memory
	 * cache if not then is downloaded from the server, save to file system and
	 * finally put into memroy cache
	 * 
	 * @param tile
	 *            The tile containing the tile URL and the tile x-y, zoomLevel
	 *            and renderer name
	 * @param cancellable
	 *            A Cancellable instance
	 * @return A Bitmap representing the tile
	 */
	public IBitmap getMapTile(final Tile tile, final Cancellable cancellable) {
		try {
			return getMapTile(tile, cancellable, null);
		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
			return this.mLoadingMapTile;
		}
	}

	/**
	 * The strategy to get Tiles is: If the tile is in memory cache, then is
	 * returned and put the first (LRU strategy) If not then the tile is
	 * searched on the FileSystem if the tile is found, is put into the memory
	 * cache if not then is downloaded from the server, save to file system and
	 * finally put into memroy cache
	 * 
	 * @param tile
	 *            The tile containing the tile URL and the tile x-y, zoomLevel
	 *            and renderer name
	 * @param cancellable
	 *            A Cancellable instance
	 * @param drawer
	 *            A TileDrawer with a bufferCanvas to draw the tiles also in
	 *            background
	 * @return A Bitmap representing the tile
	 */
	public IBitmap getMapTile(final Tile tile, final Cancellable cancellable,
			final TileDrawer drawer) {
		try {
			boolean returnLoadingTile = false;
			// long t1 = System.currentTimeMillis();
			String cacheURL = tile.getTileString();
			IBitmap ret = null;
			if (mTileCache != null)
				ret = this.mTileCache.getMapTile(cacheURL);			// long t2 = System.currentTimeMillis();
			// System.out.println("from cache: " + (t2 - t1) + " ms");
			if (ret != null) {
				mCtx.sendMessage(mLoadCallbackHandler,
						this.instantiateEvent(tile), MAPTILECACHE_SUCCESS_ID);
				if (drawer != null) {
					drawer.drawTile(tile.clone(), ret);
				}
				return ret;
			} else {
				final boolean prepend = lastZoomLevelRequested == tile.zoomLevel;
				if (lastZoomLevelRequested == -1 || prepend
						|| this.mode == TileProvider.MODE_OFFLINE) {
					returnLoadingTile = true;
					this.setViewExtent(null);
				}
				// this.lastZoomLevelRequested = tile.zoomLevel;
				// log.log(Level.FINE, "Cache failed, trying from FS.");
				try {
					// t1 = System.currentTimeMillis();
					if (this.mCtx.isSDMounted()) {
						if (!this.mFSTileProvider.getPendingQueue().contains(
								cacheURL)) {
							// log.log(Level.FINE, "load from disk");
							if (this.mFSTileProvider
									.loadMapTileToMemCacheAsync(tile,
											this.mLoadCallbackHandler,
											cacheURL, cancellable, drawer,
											prepend)) {
								if (this.mode == TileProvider.MODE_OFFLINE)
									ret = this.mOfflineMapTile;
								else
									ret = this.mLoadingMapTile;
								// t2 = System.currentTimeMillis();
								// System.out.println("from fs: " + (t2 - t1)
								// + " ms");
							} else {
								ret = this.downloadFromServer(tile, cacheURL,
										cancellable, drawer, prepend);
							}
						} else {
							if (this.mode == TileProvider.MODE_OFFLINE)
								ret = this.mOfflineMapTile;
							else
								ret = this.mLoadingMapTile;
						}
					} else {
						ret = this.downloadFromServer(tile, cacheURL,
								cancellable, drawer, prepend);
					}
				} catch (Exception e) {
					ret = this.downloadFromServer(tile, cacheURL, cancellable,
							drawer, prepend);
					e.printStackTrace();

					// log.log(Level.SEVERE, e);
				}
				// if (ret == null) {
				//
				// }
			}
			// if (ret == this.mLoadingMapTile) return null;
			if (returnLoadingTile)
				return ret;
			else
				return null;

		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
			return this.mLoadingMapTile;
		}
	}

	public IBitmap downloadFromServer(final Tile tile, final String cacheURL,
			final Cancellable cancellable, final TileDrawer drawer,
			final boolean prepend) {
		// long t1 = System.currentTimeMillis();
		if (TileProvider.this.mode != TileProvider.MODE_OFFLINE
				&& TileProvider.this.mode != TileProvider.MODE_DELETE) {
			// log.log(Level.FINE, "load from server");
			// System.out.println("4");
			this.mTileDownloader.requestMapTileAsync(tile,
					this.mLoadCallbackHandler, cacheURL, cancellable, drawer,
					prepend);
			// long t2 = System.currentTimeMillis();
			// System.out.println("from server: " + (t2 - t1) + " ms");
			return this.mLoadingMapTile;
		} else {
			mCtx.sendMessage(mLoadCallbackHandler,
					this.mTileDownloader.instantiateEvent(tile),
					MAPTILEDOWNLOADER_SUCCESS_ID);

			if (drawer != null)
				drawer.drawTile(tile.clone(), this.mOfflineMapTile);
			// log.log(Level.FINE, "Loaded from server: ");
			// mMapTileFSProvider.datas = null;
			this.mTileDownloader.mPending.remove(cacheURL);
			if (this.mFSTileProvider.getPendingQueue().contains(cacheURL))
				return null;
			return this.mOfflineMapTile;
		}
	}

	/**
	 * Clears the TileCache instance
	 */
	public void onLowMemory() {
		try {
			if (this.mTileCache != null)
				this.mTileCache.onLowMemory();
		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
		}
	}

	/**
	 * Frees memory
	 */
	public void destroy() {
		try {
			mLoadingMapTile = null;
			if (mTileCache != null)
				mTileCache.destroy();
			mTileCache = null;
			mFSTileProvider.destroy();
			mFSTileProvider = null;
			mTileDownloader = null;
		} catch (Exception e) {
			log.log(Level.SEVERE, "", e);
		}
	}

	public ITileFileSystemProvider getMFSTileProvider() {
		return mFSTileProvider;
	}

	public void setMFSTileProvider(ITileFileSystemProvider tileProvider) {
		mFSTileProvider = tileProvider;
	}

	/**
	 * @see TileProvider.MODE_*
	 * @return
	 */
	public int getMode() {
		return mode;
	}

	public void setMode(int mode) {
		this.mode = mode;
		this.getMFSTileProvider().setMode(mode);
	}

	protected TileEvent instantiateEvent(Tile aTile) {
		return new TileEvent(aTile.clone(), "", this.getMode(),
				this.mFSTileProvider.getStrategy().getName(),
				TileEvent.TILE_LEVEL_FILE_SYSTEM);
	}

	public void setLastZoomLevelRequested(int lastZoomLevelRequested) {
		this.lastZoomLevelRequested = lastZoomLevelRequested;
	}

	public void setViewExtent(Extent viewExtent) {
		this.viewExtent = viewExtent;
		this.mTileDownloader.setViewExtent(viewExtent);
	}

	public Downloader getDownloader() {
		return this.mTileDownloader;
	}
}
